# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re

from awscli.autocomplete import LazyClientCreator
from awscli.autocomplete.completer import BaseCompleter, CompletionResult
from awscli.autocomplete.filters import startswith_filter


def strip_html_tags_and_newlines(text):
    clean = re.compile('<.*?>')
    return re.sub(clean, '', text).replace('\n', '')


class ProfileCompleter(BaseCompleter):
    def __init__(
        self,
        session=None,
        response_filter=startswith_filter,
        session_creator=None,
    ):
        self._session = session
        self._filter = response_filter
        self._session_creator = session_creator
        if self._session_creator is None:
            self._session_creator = LazyClientCreator()

    def complete(self, parsed):
        if (
            parsed.current_param == 'profile'
            and parsed.current_fragment is not None
        ):
            return self._filter(parsed.current_fragment, self._get_profiles())

    def _get_profiles(self):
        return map(
            lambda x: CompletionResult(name=x),
            self._get_session().available_profiles,
        )

    def _get_session(self):
        if self._session is None:
            self._session = self._session_creator.create_session()
        return self._session


class RegionCompleter(BaseCompleter):
    def __init__(
        self,
        session=None,
        response_filter=startswith_filter,
        session_creator=None,
    ):
        self._session = session
        self._filter = response_filter
        self._session_creator = session_creator
        if self._session_creator is None:
            self._session_creator = LazyClientCreator()

    def complete(self, parsed):
        if (
            parsed.current_param == 'region'
            and parsed.current_fragment is not None
        ):
            if len(parsed.lineage) > 1:
                service_name = parsed.lineage[1]
            else:
                service_name = 'ec2'
            return self._filter(
                parsed.current_fragment,
                self._get_region_completions(service_name),
            )

    def _get_region_completions(self, service_name):
        return map(
            lambda x: CompletionResult(name=x), self._get_regions(service_name)
        )

    def _get_regions(self, service_name):
        if self._session is None:
            self._session = self._session_creator.create_session()
        return self._session.get_available_regions(service_name=service_name)


class FilePathCompleter(BaseCompleter):
    def __init__(self, path_completer=None, response_filter=startswith_filter):
        self._path_completer = path_completer
        self._filter = response_filter

    @property
    def path_completer(self):
        if self._path_completer is None:
            from prompt_toolkit.completion import PathCompleter

            self._path_completer = PathCompleter(expanduser=True)
        return self._path_completer

    def complete(self, parsed):
        if parsed.current_fragment and parsed.current_fragment.startswith(
            ('file://', 'fileb://')
        ):
            from prompt_toolkit.document import Document

            for prefix in ['file://', 'fileb://']:
                if parsed.current_fragment.startswith(prefix):
                    filename_part = parsed.current_fragment[len(prefix) :]
                    break
            # PathCompleter makes really strange suggestions for lonely ~
            # "username/", so in this case we handle it by ourselves
            if filename_part == '~':
                return [CompletionResult(''.join([prefix, '~%s' % os.sep]))]
            dirname = os.path.dirname(filename_part)
            if dirname and dirname != os.sep:
                dirname = f'{dirname}{os.sep}'
            document = Document(text=dirname, cursor_position=len(dirname))
            completions = self.path_completer.get_completions(document, None)
            results = [
                CompletionResult(
                    f'{prefix}'
                    f'{os.path.join(dirname, completion.display[0][1])}',
                    display_text=completion.display[0][1],
                )
                for completion in completions
            ]
            return self._filter(os.path.basename(filename_part), results)


class ModelIndexCompleter(BaseCompleter):
    def __init__(
        self, index, cli_driver_fetcher=None, response_filter=startswith_filter
    ):
        self._index = index
        self._cli_driver_fetcher = cli_driver_fetcher
        self._filter = response_filter

    def complete(self, parsed):
        are_unparsed_items_paths = [
            bool(re.search('[./\\\\:]|(--)', item))
            for item in parsed.unparsed_items
        ]
        if parsed.unparsed_items and all(are_unparsed_items_paths):
            # If all the unparsed items are file paths, then we auto-complete
            # options for the current fragment. This is to provide
            # auto-completion options for commands that take file paths, such
            # as `aws s3 mv ./path/to/file1 s3://path/to/file2`. We make an
            # exception for the last unparsed item if it starts with '--',
            # which indicates that the user is looking to complete an --option.
            #
            # Note that parsed.current_fragment may contain an empty string, so
            # we provide auto-completion options for the current_command
            # instead.
            if not parsed.current_fragment:
                parsed.current_fragment = parsed.current_command
            return self._filter(
                parsed.current_fragment, self._complete_options(parsed)
            )

        elif (
            parsed.unparsed_items
            or parsed.current_fragment is None
            or parsed.current_param
        ):
            # If there's ever any unparsed items, then the parser
            # encountered something it didn't understand.  We won't
            # attempt to auto-complete anything here.
            return
        current_fragment = parsed.current_fragment
        if current_fragment.startswith('--'):
            # We could technically offer completion of options
            # if the last fragment is an empty string, but to avoid
            # dumping too much information back to the user, we only
            # offer completions for options if the value starts with
            # '--'.
            prefix = parsed.current_fragment
            if prefix == '--':
                prefix = ''
            return self._filter(prefix, self._complete_options(parsed))
        else:
            # We only offer completion of options if there are no
            # more commands to complete.
            commands = self._complete_command(parsed)
            if not commands:
                return self._filter(
                    parsed.current_fragment, self._complete_options(parsed)
                )
            return self._filter(parsed.current_fragment, commands)

    def _complete_command(self, parsed):
        lineage = parsed.lineage + [parsed.current_command]
        offset = -len(parsed.current_fragment)
        result = [
            CompletionResult(name, help_text=full_name, starting_index=offset)
            for name, full_name in self._index.commands_with_full_name(lineage)
        ]
        return result

    def _outfile_filter(self, completion):
        if completion.name == '--outfile':
            completion.name = 'outfile'
            completion.display_text = 'outfile (required)'
        return completion

    def _complete_options(self, parsed):
        # '--endpoint' -> 'endpoint'
        offset = -len(parsed.current_fragment)
        is_in_global_scope = (
            parsed.lineage == [] and parsed.current_command == 'aws'
        )
        arg_names = self._index.arg_names(
            lineage=parsed.lineage, command_name=parsed.current_command
        )
        results = []
        if not is_in_global_scope:
            for arg_name in arg_names:
                arg_data = self._index.get_argument_data(
                    lineage=parsed.lineage,
                    command_name=parsed.current_command,
                    arg_name=arg_name,
                )
                help_text = None
                if self._cli_driver_fetcher:
                    help_text = strip_html_tags_and_newlines(
                        self._cli_driver_fetcher.get_argument_documentation(
                            parsed.lineage, parsed.current_command, arg_name
                        )
                    )
                results.append(
                    self._outfile_filter(
                        CompletionResult(
                            '--%s' % arg_name,
                            starting_index=offset,
                            required=arg_data.required,
                            cli_type_name=arg_data.type_name,
                            help_text=help_text,
                        )
                    )
                )
        # Global params apply to any scope
        self._inject_global_params(parsed, results)
        return [
            result
            for result in results
            if result.name.strip('--')
            not in (list(parsed.parsed_params) + list(parsed.global_params))
        ]

    def _inject_global_params(self, parsed, results):
        offset = -len(parsed.current_fragment)
        arg_data = self._index.get_global_arg_data()
        global_param_completions = []
        for arg_name, type_name, *_, help_text in arg_data:
            help_text = None
            if self._cli_driver_fetcher:
                help_text = strip_html_tags_and_newlines(
                    self._cli_driver_fetcher.get_global_arg_documentation(
                        arg_name
                    )
                )
            global_param_completions.append(
                CompletionResult(
                    '--%s' % arg_name,
                    starting_index=offset,
                    required=False,
                    cli_type_name=type_name,
                    help_text=help_text,
                )
            )
        results.extend(global_param_completions)


class ShorthandCompleter(BaseCompleter):
    _PARENS = {"[": "]", "{": "}"}
    _DUMMY_KEY_VALUE = 'cli_placeholder_for_key_value_replacement'
    _DUMMY_VALUE = 'cli_placeholder_for_value_replacement'
    _DUMMY_EQ_VALUE = 'cli_placeholder_for_key_eq_replacement'
    _VALUE_PREFIXES = {'structure': '{', 'list': '[', 'map': '{'}

    def __init__(
        self, cli_driver_fetcher=None, response_filter=startswith_filter
    ):
        self._filter = response_filter
        self._cli_driver_fetcher = cli_driver_fetcher
        self._shorthand_parser = None

    @property
    def shorthand_parser(self):
        if self._shorthand_parser is None:
            from awscli.shorthand import ShorthandParser

            self._shorthand_parser = ShorthandParser()
        return self._shorthand_parser

    def complete(self, parsed):
        if parsed.current_param and self._cli_driver_fetcher:
            results = []
            arg_model = self._cli_driver_fetcher.get_argument_model(
                parsed.lineage, parsed.current_command, parsed.current_param
            )
            if arg_model is None:
                results = self._get_prompt_for_global_arg(
                    parsed.current_param, parsed.current_fragment
                )
                return results
            parsed_input = self._parse_fragment(parsed.current_fragment)
            if parsed_input is not None:
                results = self._get_completion(arg_model, parsed_input)
                # prefix all the completions name property with the
                # current_fragment to make correct insert into the user
                # input line
                fragment = parsed.current_fragment or ''
                results = self._set_results_name(results, fragment)
            brackets_completion = self._get_close_brackets_completion(
                parsed.current_fragment
            )
            if brackets_completion is not None:
                results.append(brackets_completion)
            return results or None

    def _get_prompt_for_global_arg(self, arg_name, prefix):
        choices = self._cli_driver_fetcher.get_global_arg_choices(arg_name)
        if choices and prefix is not None:
            results = self._filter(
                prefix,
                [
                    CompletionResult(prefix, display_text=choice)
                    for choice in choices
                ],
            )
            return self._set_results_name(results, prefix)

    def _set_results_name(self, results, fragment):
        for result in results:
            name_part_len = len(fragment) - len(result.name)
            result.name = "%s%s" % (
                fragment[:name_part_len],
                result.display_text,
            )
        return results

    def _close_brackets(self, fragment):
        # If there any unclosed brackets in the text we try to close them
        # and we return part with closing brackets if they are "closable"
        stack = []
        for char in fragment:
            if char in self._PARENS.keys():
                stack.append(char)
            elif char in self._PARENS.values():
                if stack and self._PARENS[stack[-1]] == char:
                    stack.pop()
                else:
                    return ''
        return ''.join(self._PARENS[paren] for paren in reversed(stack))

    def _parse_fragment(self, fragment, attempt=1):
        # Try complete fragment to make it parsable and we handle such cases
        #
        # you just started typing input it will return a dict with input as key
        # --option foo -> {'foo': ''}
        #
        # you typed a comma
        # --option foo=1, -> foo=1,DUMMY_KEY_VALUE=DUMMY_KEY_VALUE
        #
        # you typed equal sign
        # --option foo= -> foo=DUMMY_VALUE
        #
        # after that we close all the brackets if needed and try to parse
        # if we get a parsing error it can be that you have some partially
        # entered nested structure in such a case we'll try to complete it
        # and make one more attempt
        # --option foo={bar -> foo={bar=DUMMY_EQ_VALUE
        from awscli.shorthand import ShorthandParseError

        if fragment is None:
            return None
        if fragment == '':
            return {}
        if '=' not in fragment:
            return {fragment: None}
        if fragment.endswith(','):
            fragment += f'{self._DUMMY_KEY_VALUE}={self._DUMMY_KEY_VALUE}'
        elif fragment.endswith('='):
            fragment += self._DUMMY_VALUE
        try:
            return self.shorthand_parser.parse(
                f'{fragment}{self._close_brackets(fragment)}'
            )
        except ShorthandParseError:
            if attempt == 1:
                return self._parse_fragment(
                    f'{fragment}={self._DUMMY_EQ_VALUE}', attempt + 1
                )
        # if we get here it means that we can't make it parsable and can't
        # suggest anything so the only solution is to wait till user enter more

    def _get_completion(self, arg_model, parsed_input):
        completion = getattr(
            self, "_get_prompt_for_%s" % arg_model.type_name, self._no_prompt
        )(arg_model, parsed_input)
        if completion is None:
            return []
        return completion

    def _get_last_key_value(self, parsed_input):
        # returns last key and its value in top level structure
        # from parsed_input
        # - if value is not None the next step will be to autopropmt value
        # if possible, value='' - means that user typed '=' but hasn't
        # started typing value yet
        # - if value is None but key is not None the next step will be to
        # autopropmt key if possible, key='' - means that user
        # typed ',' after previous key=value pair but hasn't started typing
        # key yet
        # - if both None - it means it's not a dict (maybe string or integer)
        # and we won't prompt anything
        if not isinstance(parsed_input, dict):
            return None, None
        if not parsed_input:
            return '', None
        key = list(parsed_input)[-1]
        value = parsed_input[key]
        # if get in situation when foo=a,bar it will parsed as
        # foo=['a','bar'] and we will assume that "bar" is the next key
        if value and isinstance(value, list):
            key = value[-1]
            value = None
        # remove dummy_parts if they were added before
        elif value == key == self._DUMMY_KEY_VALUE:
            key = ''
            value = None
        elif value == self._DUMMY_EQ_VALUE:
            value = None
        return key, value

    def _no_prompt(self, *args, **kwargs):
        # for all types we do not support
        return None

    def _get_close_brackets_completion(self, fragment):
        if not fragment:
            return
        close_brackets = self._close_brackets(fragment)
        if close_brackets:
            return CompletionResult(
                f'{fragment}{close_brackets}',
                display_text='Autoclose brackets',
            )

    def _get_prompt_for_string(self, arg_model, parsed_input):
        # If string shape is a value for a list of enums it'll be an instance
        # of dict because parsing method tries to convert it to a valid
        # shorthand expression. In such a case we need to take the first and
        # only key of the dict if it has keys or an empty string
        if getattr(arg_model, 'enum', False):
            prefix = parsed_input or ''
            if isinstance(prefix, dict):
                prefix = list(prefix.keys())[0]
            if prefix == self._DUMMY_VALUE:
                prefix = ''
            return self._filter(
                prefix,
                [
                    CompletionResult(prefix, display_text=enum)
                    for enum in arg_model.enum
                ],
            )

    def _get_prompt_for_boolean(self, arg_model, parsed_input):
        all_results = ['true', 'false']
        prefix = parsed_input
        if prefix == self._DUMMY_VALUE:
            prefix = ''
        return self._filter(
            prefix,
            [
                CompletionResult(prefix, display_text=result)
                for result in all_results
            ],
        )

    def _get_prompt_for_list(self, arg_model, parsed_input):
        # we have two way we can enter lists:
        # - if it's a top level structure it will be space separated and
        # parsed_input will contain the last element of the list
        # - if it's a nested structure we'll get the whole list as an input
        # and take the last item
        # - if list is empty we'll pass None to the next completer
        if isinstance(parsed_input, list):
            if parsed_input:
                parsed_input = parsed_input[-1]
            else:
                parsed_input = None
        return self._get_completion(arg_model.member, parsed_input)

    def _get_prompt_for_map(self, arg_model, parsed_input):
        last_key, last_value = self._get_last_key_value(parsed_input)
        if last_value is not None:
            return self._get_completion(arg_model.value, last_value)
        return self._get_completion(arg_model.key, last_key)

    def _get_prompt_for_structure(self, arg_model, parsed_input):
        last_key, last_value = self._get_last_key_value(parsed_input)
        if last_key is None and last_value is None:
            return None
        # if we have a last_value it can be nested structure so let's try
        # to get suggestion for it
        if last_key and last_value is not None:
            if not arg_model.members.get(last_key):
                # key exists but we don't have such key in model
                return None
            return self._get_completion(
                arg_model.members[last_key], last_value
            )
        entered_keys = set(parsed_input) - set([last_key])
        return self._get_struct_keys_completions(
            arg_model, entered_keys, last_key
        )

    def _get_struct_keys_completions(self, arg_model, entered_keys, last_key):
        # get suggestions for the structure keys, as CompletionResult.name
        # we return only part of the suggestion that has not been entered yet
        results = []
        for member_name, member in arg_model.members.items():
            if member_name not in entered_keys:
                display_text = '%s=%s' % (
                    member_name,
                    self._VALUE_PREFIXES.get(member.type_name, ''),
                )
                results.append(
                    CompletionResult(
                        last_key,
                        help_text=strip_html_tags_and_newlines(
                            member.documentation
                        ),
                        cli_type_name=member.type_name,
                        display_text=display_text,
                    )
                )
        return self._filter(last_key, results)


class QueryCompleter(BaseCompleter):
    def __init__(
        self, cli_driver_fetcher=None, response_filter=startswith_filter
    ):
        self._filter = response_filter
        self._cli_driver_fetcher = cli_driver_fetcher
        self._argument_generator = None
        self._jmespath = None

    @property
    def jmespath(self):
        if self._jmespath is None:
            import jmespath

            self._jmespath = jmespath
        return self._jmespath

    @property
    def argument_generator(self):
        if self._argument_generator is None:
            from botocore.utils import ArgumentGenerator

            self._argument_generator = ArgumentGenerator
        return self._argument_generator

    def complete(self, parsed):
        if self._cli_driver_fetcher is None:
            return
        if (
            parsed.current_param == 'query'
            and parsed.current_fragment is not None
        ):
            operation_model = self._cli_driver_fetcher.get_operation_model(
                parsed.lineage, parsed.current_command
            )
            if operation_model:
                return self._get_completions(
                    parsed.current_fragment, operation_model
                )

    def _get_query_and_last_key(self, query):
        # Because output example has only 1 element in any list if
        # user enters query like Groups[2] jmespath will return "null"
        # and we won't be able to prompting so we change all the numbers
        # in brackets to 0 to be able to get the shape of the nested structure
        query = re.sub(r'([\{\[])\d+?([\}\]])', r'\g<1>0\g<2>', query)
        return query.rsplit('.', 1)

    def _create_completions(self, results, last_key, fragment):
        completions = self._filter(
            last_key,
            [
                CompletionResult(last_key, display_text=result)
                for result in results
            ],
        )
        for completion in completions:
            name_part_len = len(fragment) - len(completion.name)
            completion.name = "%s%s" % (
                fragment[:name_part_len],
                completion.display_text,
            )
        return completions

    def _is_field_expression(self, expression):
        is_last_child_field = False
        is_field = expression.parsed['type'] == 'field'
        if expression.parsed['children']:
            is_last_child_field = (
                expression.parsed['children'][-1]['type'] == 'field'
            )
        return is_field or is_last_child_field

    def _get_completions(self, fragment, operation_model):
        results = []
        last_key = fragment
        if operation_model.output_shape:
            argument_generator = self.argument_generator(use_member_names=True)
            response = argument_generator.generate_skeleton(
                operation_model.output_shape
            )
            if '.' not in fragment:
                if isinstance(response, dict):
                    results = response.keys()
            else:
                try:
                    query, last_key = self._get_query_and_last_key(fragment)
                    expression = self.jmespath.compile(query)
                    parsed_response = expression.search(response)
                    if (
                        parsed_response
                        and isinstance(parsed_response, list)
                        and not self._is_field_expression(expression)
                    ):
                        parsed_response = parsed_response[0]
                    if isinstance(parsed_response, dict):
                        results = parsed_response.keys()
                except self.jmespath.exceptions.JMESPathError:
                    pass
        return self._create_completions(results, last_key, fragment)



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)

logger = logging.getLogger(__name__)


class FipsEndpointUnsupported(ConfigurationError):
    pass


class InstanceConnectEndpointRequestFetcher:
    def get_eice_dns_name(self, eice_info, is_fips_enabled):
        fips_dns_name = eice_info.get('FipsDnsName')

        if not is_fips_enabled:
            return eice_info['DnsName']
        elif is_fips_enabled and fips_dns_name:
            return fips_dns_name
        elif is_fips_enabled and not fips_dns_name:
            raise FipsEndpointUnsupported("Unable to find FIPS Endpoint")

    def get_available_instance_connect_endpoint(
        self, ec2_client, vpc_id, subnet_id, instance_connect_endpoint_id
    ):
        if instance_connect_endpoint_id:
            return self._get_instance_connect_endpoint_by_id(
                ec2_client, instance_connect_endpoint_id
            )
        else:
            return self._get_instance_connect_endpoint_by_vpc(
                ec2_client, vpc_id, subnet_id
            )

    def _get_instance_connect_endpoint_by_id(
        self, ec2_client, instance_connect_endpoint_id
    ):
        args = {
            "Filters": [
                {
                    "Name": "state",
                    "Values": [
                        "create-complete",
                        "update-in-progress",
                        "update-failed",
                        "update-complete",
                    ],
                }
            ],
            "InstanceConnectEndpointIds": [instance_connect_endpoint_id],
        }
        describe_eice_response = (
            ec2_client.describe_instance_connect_endpoints(**args)
        )
        instance_connect_endpoints = describe_eice_response[
            "InstanceConnectEndpoints"
        ]
        if instance_connect_endpoints:
            return instance_connect_endpoints[0]
        raise ParamValidationError(
            f"There are no available instance connect endpoints with {instance_connect_endpoint_id}"
        )

    def _get_instance_connect_endpoint_by_vpc(
        self, ec2_client, vpc_id, subnet_id
    ):
        ## Describe until subnet match and if none match subnet then return the first one based on vpc-id filter
        args = {
            "Filters": [
                {
                    "Name": "state",
                    "Values": [
                        "create-complete",
                        "update-in-progress",
                        "update-failed",
                        "update-complete",
                    ],
                },
                {"Name": "vpc-id", "Values": [vpc_id]},
            ]
        }

        paginator = ec2_client.get_paginator(
            'describe_instance_connect_endpoints'
        )
        page_iterator = paginator.paginate(**args)
        instance_connect_endpoints = []
        for page in page_iterator:
            page_result = page["InstanceConnectEndpoints"]
            instance_connect_endpoints = (
                instance_connect_endpoints + page_result
            )
            if page_result:
                for eice in page_result:
                    if eice['SubnetId'] == subnet_id:
                        logger.debug(f"Using EICE based on subnet: {eice}")
                        return eice

        if instance_connect_endpoints:
            logger.debug(
                f"Using EICE based on vpc: {instance_connect_endpoints[0]}"
            )
            return instance_connect_endpoints[0]

        raise ParamValidationError(
            "There are no available instance connect endpoints."
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import sys

from awscli.customizations.commands import BasicCommand
from awscli.customizations.utils import validate_mutually_exclusive

from . import (
    PREDEFINED_SECTION_NAMES,
    SUBSECTION_TYPE_ALLOWLIST,
    SubsectionNotFoundError,
)

LOG = logging.getLogger(__name__)


class ConfigureGetCommand(BasicCommand):
    NAME = 'get'
    DESCRIPTION = BasicCommand.FROM_FILE(
        'configure', 'get', '_description.rst'
    )
    SYNOPSIS = 'aws configure get varname [--profile profile-name]'
    EXAMPLES = BasicCommand.FROM_FILE('configure', 'get', '_examples.rst')
    ARG_TABLE = [
        {
            'name': 'varname',
            'help_text': 'The name of the config value to retrieve.',
            'action': 'store',
            'cli_type_name': 'string',
            'positional_arg': True,
        },
        {
            'name': 'sso-session',
            'help_text': 'The name of the sub-section from which to retrieve a value.',
            'action': 'store',
            'cli_type_name': 'string',
        },
        {
            'name': 'services',
            'help_text': 'The name of the sub-section from which to retrieve a value.',
            'action': 'store',
            'cli_type_name': 'string',
        },
    ]

    def __init__(self, session, stream=None, error_stream=None):
        super().__init__(session)
        if stream is None:
            stream = sys.stdout
        if error_stream is None:
            error_stream = sys.stderr
        self._stream = stream
        self._error_stream = error_stream

    def _subsection_parameter_to_argument_name(self, parameter_name):
        return parameter_name.replace("-", "_")

    def _get_subsection_config_name_from_args(self, args):
        # Validate mutual exclusivity of sub-section type parameters
        groups = [
            [self._subsection_parameter_to_argument_name(key)]
            for key in SUBSECTION_TYPE_ALLOWLIST.keys()
        ]
        validate_mutually_exclusive(args, *groups)

        subsection_name = None
        subsection_config_name = None

        for (
            section_type,
            section_properties,
        ) in SUBSECTION_TYPE_ALLOWLIST.items():
            cli_parameter_name = self._subsection_parameter_to_argument_name(
                section_type
            )

            if hasattr(args, cli_parameter_name):
                subsection_name = getattr(args, cli_parameter_name)
                if subsection_name is not None:
                    subsection_config_name = section_properties[
                        'full_config_name'
                    ]
                    break

        return (subsection_config_name, subsection_name)

    def _run_main(self, args, parsed_globals):
        varname = args.varname
        section_type, section_name = (
            self._get_subsection_config_name_from_args(args)
        )

        if section_type:
            value = self._get_subsection_property(
                section_type.replace("-", "_"), section_name, varname
            )
        elif '.' not in varname:
            # get_scoped_config() returns the config variables in the config
            # file (not the logical_var names), which is what we want.
            config = self._session.get_scoped_config()
            value = config.get(varname)
        else:
            value = self._get_dotted_config_value(varname)

        LOG.debug(f'Config value retrieved: {value}')

        if isinstance(value, str):
            self._stream.write(value)
            self._stream.write('\n')
            return 0
        elif isinstance(value, dict):
            # TODO: add support for this. We would need to print it off in
            # the same format as the config file.
            self._error_stream.write(
                f'varname ({varname}) must reference a value, not a section '
                'or sub-section.'
            )
            return 1
        else:
            return 1

    def _get_dotted_config_value(self, varname):
        parts = varname.split('.')
        num_dots = varname.count('.')

        # Logic to deal with predefined sections like [plugin] etc.
        if num_dots == 1 and parts[0] in PREDEFINED_SECTION_NAMES:
            full_config = self._session.full_config
            section, config_name = varname.split('.')
            value = full_config.get(section, {}).get(config_name)
            if value is None:
                # Try to retrieve it from the profile config.
                value = (
                    full_config['profiles'].get(section, {}).get(config_name)
                )
            return value

        if parts[0] == 'profile':
            profile_name = parts[1]
            config_name = parts[2]
            remaining = parts[3:]
        # Check if varname starts with 'default' profile (e.g.
        # default.emr-dev.emr.instance_profile) If not, go further to check
        # if varname starts with a known profile name
        elif parts[0] == 'default' or (
            parts[0] in self._session.full_config['profiles']
        ):
            profile_name = parts[0]
            config_name = parts[1]
            remaining = parts[2:]
        else:
            profile_name = self._session.get_config_variable('profile')
            if profile_name is None:
                profile_name = 'default'
            config_name = parts[0]
            remaining = parts[1:]

        value = (
            self._session.full_config['profiles']
            .get(profile_name, {})
            .get(config_name)
        )
        if len(remaining) == 1:
            try:
                value = value.get(remaining[-1])
            except AttributeError:
                value = None
        return value

    def _get_subsection_property(self, section_type, section_name, varname):
        full_config = self._session.full_config

        if section_type not in full_config:
            raise SubsectionNotFoundError(
                f"The config sub-section ({section_name}) could not be found"
            )

        section_type_config = full_config[section_type]
        section_config = section_type_config.get(section_name, None)

        if section_config is None:
            raise SubsectionNotFoundError(
                f"The config sub-section ({section_name}) could not be found"
            )

        # Handle nested properties
        if '.' in varname:
            parts = varname.split('.')
            if len(parts) == 2:
                value = section_config.get(parts[0])
                if isinstance(value, dict):
                    return value.get(parts[1])
                else:
                    return None
        else:
            return section_config.get(varname)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import platform
import re
from socket import timeout

import awscli.compat
from awscli.compat import URLError, urlopen
from awscli.customizations.codedeploy.systems import (
    RHEL,
    System,
    Ubuntu,
    Windows,
)
from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)

MAX_INSTANCE_NAME_LENGTH = 100
MAX_TAGS_PER_INSTANCE = 10
MAX_TAG_KEY_LENGTH = 128
MAX_TAG_VALUE_LENGTH = 256

INSTANCE_NAME_PATTERN = r'^[A-Za-z0-9+=,.@_-]+$'
IAM_USER_ARN_PATTERN = r'^arn:aws:iam::[0-9]{12}:user/[A-Za-z0-9/+=,.@_-]+$'

INSTANCE_NAME_ARG = {
    'name': 'instance-name',
    'synopsis': '--instance-name <instance-name>',
    'required': True,
    'help_text': ('Required. The name of the on-premises instance.'),
}

IAM_USER_ARN_ARG = {
    'name': 'iam-user-arn',
    'synopsis': '--iam-user-arn <iam-user-arn>',
    'required': False,
    'help_text': (
        'Optional. The IAM user associated with the on-premises instance.'
    ),
}


def validate_region(params, parsed_globals):
    if parsed_globals.region:
        params.region = parsed_globals.region
    else:
        params.region = params.session.get_config_variable('region')
    if not params.region:
        raise ConfigurationError('Region not specified.')


def validate_instance_name(params):
    if params.instance_name:
        if not re.match(INSTANCE_NAME_PATTERN, params.instance_name):
            raise ParamValidationError(
                'Instance name contains invalid characters.'
            )
        if params.instance_name.startswith('i-'):
            raise ParamValidationError(
                'Instance name cannot start with \'i-\'.'
            )
        if len(params.instance_name) > MAX_INSTANCE_NAME_LENGTH:
            raise ParamValidationError(
                f'Instance name cannot be longer than {MAX_INSTANCE_NAME_LENGTH} characters.'
            )


def validate_tags(params):
    if params.tags:
        if len(params.tags) > MAX_TAGS_PER_INSTANCE:
            raise ParamValidationError(
                f'Instances can only have a maximum of {MAX_TAGS_PER_INSTANCE} tags.'
            )
        for tag in params.tags:
            if len(tag['Key']) > MAX_TAG_KEY_LENGTH:
                raise ParamValidationError(
                    f'Tag Key cannot be longer than {MAX_TAG_KEY_LENGTH} characters.'
                )
            if len(tag['Value']) > MAX_TAG_VALUE_LENGTH:
                raise ParamValidationError(
                    f'Tag Value cannot be longer than {MAX_TAG_VALUE_LENGTH} characters.'
                )


def validate_iam_user_arn(params):
    if params.iam_user_arn and not re.match(
        IAM_USER_ARN_PATTERN, params.iam_user_arn
    ):
        raise ParamValidationError('Invalid IAM user ARN.')


def validate_instance(params):
    if platform.system() == 'Linux':
        distribution = awscli.compat.linux_distribution()[0]
        if 'Ubuntu' in distribution:
            params.system = Ubuntu(params)
        if 'Red Hat Enterprise Linux Server' in distribution:
            params.system = RHEL(params)
    elif platform.system() == 'Windows':
        params.system = Windows(params)
    if 'system' not in params:
        raise RuntimeError(System.UNSUPPORTED_SYSTEM_MSG)
    try:
        urlopen('http://169.254.169.254/latest/meta-data/', timeout=1)
        raise RuntimeError('Amazon EC2 instances are not supported.')
    except (URLError, timeout):
        pass


def validate_s3_location(params, arg_name):
    arg_name = arg_name.replace('-', '_')
    if arg_name in params:
        s3_location = getattr(params, arg_name)
        if s3_location:
            matcher = re.match('s3://(.+?)/(.+)', str(s3_location))
            if matcher:
                params.bucket = matcher.group(1)
                params.key = matcher.group(2)
            else:
                raise ParamValidationError(
                    '--{0} must specify the Amazon S3 URL format as '
                    's3://<bucket>/<key>.'.format(arg_name.replace('_', '-'))
                )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Stream display implementation for ECS Express Gateway Service monitoring."""

import time
from datetime import datetime, timezone

from awscli.customizations.ecs.expressgateway.managedresourcegroup import (
    ManagedResourceGroup,
)
from awscli.customizations.utils import uni_print


class StreamDisplay:
    """Stream display for monitoring that outputs changes to stdout.

    Provides text-based monitoring output suitable for non-interactive
    environments, logging, or piping to other commands.
    """

    def __init__(self, use_color=True):
        self.previous_resources_by_key = {}
        self.use_color = use_color

    def show_startup_message(self):
        """Show startup message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Starting monitoring...\n")

    def show_polling_message(self):
        """Show polling message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Polling for updates...\n")

    def show_monitoring_data(self, resource_group, info):
        """Show monitoring data for resources with diff detection.

        Args:
            resource_group: ManagedResourceGroup or None
            info: Additional info text to display
        """
        timestamp = self._get_timestamp()

        if resource_group:
            (
                changed_resources,
                updated_dict,
                removed_keys,
            ) = resource_group.get_changed_resources(
                self.previous_resources_by_key
            )

            self.previous_resources_by_key = updated_dict

            if changed_resources:
                self._print_flattened_resources_list(
                    changed_resources, timestamp
                )

        if info:
            uni_print(f"[{timestamp}] {info}\n")

    def _print_flattened_resources_list(self, resources_list, timestamp):
        """Print individual resources from a flat list as timestamped lines.

        Args:
            resources_list: List of ManagedResource objects to print
            timestamp: Timestamp string to prefix each line
        """
        for resource in resources_list:
            output = resource.get_stream_string(timestamp, self.use_color)
            uni_print(output + "\n")

    def show_timeout_message(self):
        """Show timeout message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Monitoring timeout reached!\n")

    def show_service_inactive_message(self):
        """Show service inactive message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Service is inactive\n")

    def show_completion_message(self):
        """Show completion message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Monitoring complete!\n")

    def show_user_stop_message(self):
        """Show user stop message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Monitoring stopped by user\n")

    def show_error_message(self, error):
        """Show error message."""
        timestamp = self._get_timestamp()
        uni_print(f"[{timestamp}] Error: {error}\n")

    def _get_timestamp(self):
        """Get formatted timestamp."""
        return datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%SZ")



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from traceback import format_tb


class BaseWizardException(Exception):
    pass


class InvalidDataTypeConversionException(BaseWizardException):
    MSG_FORMAT = 'Invalid value {value} for datatype {datatype}'

    def __init__(self, value, datatype):
        message = self.MSG_FORMAT.format(value=value, datatype=datatype)
        super().__init__(message)


class InvalidChoiceException(Exception):
    pass


class UnableToRunWizardError(Exception):
    pass


class UnexpectedWizardException(Exception):
    MSG_FORMAT = (
        'Encountered unexpected exception inside of wizard:\n\n'
        'Traceback:\n{original_tb}'
        '{original_exception_cls}: {original_exception}'
    )

    def __init__(self, original_exception):
        self.original_exception = original_exception
        message = self.MSG_FORMAT.format(
            original_tb=''.join(format_tb(original_exception.__traceback__)),
            original_exception_cls=self.original_exception.__class__.__name__,
            original_exception=self.original_exception,
        )
        super().__init__(message)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.styles import Style
from prompt_toolkit.utils import is_windows


def get_default_style():
    basic_styles = [
        # Wizard-specific classes
        ('wizard', ''),
        ('wizard.title', 'underline bold'),
        ('wizard.prompt.description', 'bold'),
        ('wizard.prompt.description.current', 'white'),
        ('wizard.prompt.answer', 'bg:#aaaaaa black'),
        ('wizard.prompt.answer.current', 'white'),
        ('wizard.section.tab.current', 'white'),
        ('wizard.section.tab.unvisited', '#777777'),
        ('wizard.section.tab.visited', ''),
        ('wizard.dialog', ''),
        ('wizard.dialog frame.label', 'white bold'),
        ('wizard.dialog.save frame.label', 'black bold'),
        ('wizard.dialog.body', 'bg:#aaaaaa black'),
        ('wizard.error', 'bg:#550000 #ffffff'),
        # Prompt-toolkit built-in classes
        ('button.focused', 'bg:#777777 white'),
        ('completion-menu.completion', 'underline'),
    ]
    if is_windows():
        os_related_styles = [
            ('wizard.section.tab', 'bold black'),
            ('shadow', 'bg:#eeeeee'),
        ]
    else:
        os_related_styles = [
            ('wizard.section.tab', 'bold bg:#aaaaaa black'),
            ('shadow', 'bg:#222222'),
        ]
    return Style(basic_styles + os_related_styles)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Module for processing CLI args."""

import logging
import os

from botocore.compat import OrderedDict, json
from botocore.utils import is_json_value_header

from awscli import COMPLEX_TYPES, SCALAR_TYPES, shorthand
from awscli.utils import (
    find_service_and_method_in_event_name,
    is_document_type,
    is_document_type_container,
)

LOG = logging.getLogger('awscli.argprocess')


class ParamError(Exception):
    def __init__(self, cli_name, message):
        """

        :type cli_name: string
        :param cli_name: The complete cli argument name,
            e.g. "--foo-bar".  It should include the leading
            hyphens if that's how a user would specify the name.

        :type message: string
        :param message: The error message to display to the user.

        """
        full_message = "Error parsing parameter '%s': %s" % (cli_name, message)
        super(ParamError, self).__init__(full_message)
        self.cli_name = cli_name
        self.message = message


class ParamSyntaxError(Exception):
    pass


class ParamUnknownKeyError(Exception):
    def __init__(self, key, valid_keys):
        valid_keys = ', '.join(valid_keys)
        full_message = f"Unknown key '{key}', valid choices are: {valid_keys}"
        super(ParamUnknownKeyError, self).__init__(full_message)


class TooComplexError(Exception):
    pass


def unpack_argument(
    session, service_name, operation_name, cli_argument, value
):
    """
    Unpack an argument's value from the commandline. This is part one of a two
    step process in handling commandline arguments. Emits the load-cli-arg
    event with service, operation, and parameter names. Example::

        load-cli-arg.ec2.describe-instances.foo

    """
    param_name = getattr(cli_argument, 'name', 'anonymous')

    value_override = session.emit_first_non_none_response(
        'load-cli-arg.%s.%s.%s' % (service_name, operation_name, param_name),
        param=cli_argument,
        value=value,
        service_name=service_name,
        operation_name=operation_name,
    )

    if value_override is not None:
        value = value_override

    return value


def detect_shape_structure(param):
    stack = []
    return _detect_shape_structure(param, stack)


def _detect_shape_structure(param, stack):
    if param.name in stack:
        return 'recursive'
    else:
        stack.append(param.name)
    try:
        if param.type_name in SCALAR_TYPES:
            return 'scalar'
        elif param.type_name == 'structure':
            sub_types = [
                _detect_shape_structure(p, stack)
                for p in param.members.values()
            ]
            # We're distinguishing between structure(scalar)
            # and structure(scalars), because for the case of
            # a single scalar in a structure we can simplify
            # more than a structure(scalars).
            if len(sub_types) == 1 and all(p == 'scalar' for p in sub_types):
                return 'structure(scalar)'
            elif len(sub_types) > 1 and all(p == 'scalar' for p in sub_types):
                return 'structure(scalars)'
            else:
                return 'structure(%s)' % ', '.join(sorted(set(sub_types)))
        elif param.type_name == 'list':
            return 'list-%s' % _detect_shape_structure(param.member, stack)
        elif param.type_name == 'map':
            if param.value.type_name in SCALAR_TYPES:
                return 'map-scalar'
            else:
                return 'map-%s' % _detect_shape_structure(param.value, stack)
    finally:
        stack.pop()


def unpack_cli_arg(cli_argument, value):
    """
    Parses and unpacks the encoded string command line parameter
    and returns native Python data structures that can be passed
    to the Operation.

    :type cli_argument: :class:`awscli.arguments.BaseCLIArgument`
    :param cli_argument: The CLI argument object.

    :param value: The value of the parameter.  This can be a number of
        different python types (str, list, etc).  This is the value as
        it's specified on the command line.

    :return: The "unpacked" argument than can be sent to the `Operation`
        object in python.
    """
    return _unpack_cli_arg(
        cli_argument.argument_model, value, cli_argument.cli_name
    )


def _special_type(model):
    # check if model is jsonvalue header and that value is serializable
    if (
        model.serialization.get('jsonvalue')
        and model.serialization.get('location') == 'header'
        and model.type_name == 'string'
    ):
        return True
    return False


def _unpack_cli_arg(argument_model, value, cli_name):
    if is_json_value_header(argument_model) or is_document_type(
        argument_model
    ):
        return _unpack_json_cli_arg(argument_model, value, cli_name)
    elif argument_model.type_name in SCALAR_TYPES:
        return unpack_scalar_cli_arg(argument_model, value, cli_name)
    elif argument_model.type_name in COMPLEX_TYPES:
        return _unpack_complex_cli_arg(argument_model, value, cli_name)
    else:
        return str(value)


def _unpack_json_cli_arg(argument_model, value, cli_name):
    try:
        return json.loads(value, object_pairs_hook=OrderedDict)
    except ValueError as e:
        raise ParamError(
            cli_name, "Invalid JSON: %s\nJSON received: %s" % (e, value)
        )


def _unpack_complex_cli_arg(argument_model, value, cli_name):
    type_name = argument_model.type_name
    if type_name == 'structure' or type_name == 'map':
        if value.lstrip()[0] == '{':
            return _unpack_json_cli_arg(argument_model, value, cli_name)
        raise ParamError(cli_name, "Invalid JSON:\n%s" % value)
    elif type_name == 'list':
        if isinstance(value, str):
            if value.lstrip()[0] == '[':
                return _unpack_json_cli_arg(argument_model, value, cli_name)
        elif isinstance(value, list) and len(value) == 1:
            single_value = value[0].strip()
            if single_value and single_value[0] == '[':
                return _unpack_json_cli_arg(argument_model, value[0], cli_name)
        try:
            # There's a couple of cases remaining here.
            # 1. It's possible that this is just a list of strings, i.e
            # --security-group-ids sg-1 sg-2 sg-3 => ['sg-1', 'sg-2', 'sg-3']
            # 2. It's possible this is a list of json objects:
            # --filters '{"Name": ..}' '{"Name": ...}'
            member_shape_model = argument_model.member
            return [
                _unpack_cli_arg(member_shape_model, v, cli_name) for v in value
            ]
        except (ValueError, TypeError):
            # The list params don't have a name/cli_name attached to them
            # so they will have bad error messages.  We're going to
            # attach the parent parameter to this error message to provide
            # a more helpful error message.
            raise ParamError(cli_name, value[0])


def unpack_scalar_cli_arg(argument_model, value, cli_name=''):
    # Note the cli_name is used strictly for error reporting.  It's
    # not required to use unpack_scalar_cli_arg
    if (
        argument_model.type_name == 'integer'
        or argument_model.type_name == 'long'
    ):
        return int(value)
    elif (
        argument_model.type_name == 'float'
        or argument_model.type_name == 'double'
    ):
        # TODO: losing precision on double types
        return float(value)
    elif (
        argument_model.type_name == 'blob'
        and argument_model.serialization.get('streaming')
    ):
        file_path = os.path.expandvars(value)
        file_path = os.path.expanduser(file_path)
        if not os.path.isfile(file_path):
            msg = 'Blob values must be a path to a file.'
            raise ParamError(cli_name, msg)
        return open(file_path, 'rb')
    elif argument_model.type_name == 'boolean':
        if isinstance(value, str) and value.lower() == 'false':
            return False
        return bool(value)
    else:
        return value


def _supports_shorthand_syntax(model):
    # Shorthand syntax is only supported if:
    #
    # 1. The argument is not a document type nor is a wrapper around a document
    # type (e.g. is a list of document types or a map of document types). These
    # should all be expressed as JSON input.
    #
    # 2. The argument is sufficiently complex, that is, it's base type is
    # a complex type *and* if it's a list, then it can't be a list of
    # scalar types.
    if is_document_type_container(model):
        return False
    return _is_complex_shape(model)


def _is_complex_shape(model):
    if model.type_name not in ['structure', 'list', 'map']:
        return False
    elif model.type_name == 'list':
        if model.member.type_name not in ['structure', 'list', 'map']:
            return False
    return True


class ParamShorthand:
    def _uses_old_list_case(self, command_name, operation_name, argument_name):
        """
        Determines whether a given operation for a service needs to use the
        deprecated shorthand parsing case for lists of structures that only have
        a single member.
        """
        cases = {
            'firehose': {'put-record-batch': ['records']},
            'workspaces': {
                'reboot-workspaces': ['reboot-workspace-requests'],
                'rebuild-workspaces': ['rebuild-workspace-requests'],
                'terminate-workspaces': ['terminate-workspace-requests'],
            },
            'elb': {
                'remove-tags': ['tags'],
                'describe-instance-health': ['instances'],
                'deregister-instances-from-load-balancer': ['instances'],
                'register-instances-with-load-balancer': ['instances'],
            },
        }
        cases = cases.get(command_name, {}).get(operation_name, [])
        return argument_name in cases


class ParamShorthandParser(ParamShorthand):
    def __init__(self):
        self._parser = shorthand.ShorthandParser()
        self._visitor = shorthand.BackCompatVisitor()

    def __call__(self, cli_argument, value, event_name, **kwargs):
        """Attempt to parse shorthand syntax for values.

        This is intended to be hooked up as an event handler (hence the
        **kwargs).  Given ``param`` object and its string ``value``,
        figure out if we can parse it.  If we can parse it, we return
        the parsed value (typically some sort of python dict).

        :type cli_argument: :class:`awscli.arguments.BaseCLIArgument`
        :param cli_argument: The CLI argument object.

        :type param: :class:`botocore.parameters.Parameter`
        :param param: The parameter object (includes various metadata
            about the parameter).

        :type value: str
        :param value: The value for the parameter type on the command
            line, e.g ``--foo this_value``, value would be ``"this_value"``.

        :returns: If we can parse the value we return the parsed value.
            If it looks like JSON, we return None (which tells the event
            emitter to use the default ``unpack_cli_arg`` provided that
            no other event handlers can parsed the value).  If we
            run into an error parsing the value, a ``ParamError`` will
            be raised.

        """

        if not self._should_parse_as_shorthand(cli_argument, value):
            return
        else:
            command_name, operation_name = (
                find_service_and_method_in_event_name(event_name)
            )
            return self._parse_as_shorthand(
                cli_argument, value, command_name, operation_name
            )

    def _parse_as_shorthand(
        self, cli_argument, value, command_name, operation_name
    ):
        try:
            LOG.debug("Parsing param %s as shorthand", cli_argument.cli_name)
            handled_value = self._handle_special_cases(
                cli_argument, value, command_name, operation_name
            )
            if handled_value is not None:
                return handled_value
            if isinstance(value, list):
                # Because of how we're using argparse, list shapes
                # are configured with nargs='+' which means the ``value``
                # is given to us "conveniently" as a list.  When
                # this happens we need to parse each list element
                # individually.
                parsed = [self._parser.parse(v) for v in value]
                self._visitor.visit(parsed, cli_argument.argument_model)
            else:
                # Otherwise value is just a string.
                parsed = self._parser.parse(value)
                self._visitor.visit(parsed, cli_argument.argument_model)
        except shorthand.ShorthandParseError as e:
            raise ParamError(cli_argument.cli_name, str(e))
        except (ParamError, ParamUnknownKeyError) as e:
            # The shorthand parse methods don't have the cli_name,
            # so any ParamError won't have this value.  To accommodate
            # this, ParamErrors are caught and reraised with the cli_name
            # injected.
            raise ParamError(cli_argument.cli_name, str(e))
        return parsed

    def _handle_special_cases(
        self, cli_argument, value, command_name, operation_name
    ):
        # We need to handle a few special cases that the previous
        # parser handled in order to stay backwards compatible.
        model = cli_argument.argument_model
        if (
            model.type_name == 'list'
            and model.member.type_name == 'structure'
            and len(model.member.members) == 1
            and self._uses_old_list_case(
                command_name, operation_name, cli_argument.name
            )
        ):
            # First special case is handling a list of structures
            # of a single element such as:
            #
            # --instance-ids id-1 id-2 id-3
            #
            # gets parsed as:
            #
            # [{"InstanceId": "id-1"}, {"InstanceId": "id-2"},
            #  {"InstanceId": "id-3"}]
            key_name = list(model.member.members.keys())[0]
            new_values = [{key_name: v} for v in value]
            return new_values
        elif (
            model.type_name == 'structure'
            and len(model.members) == 1
            and 'Value' in model.members
            and model.members['Value'].type_name == 'string'
            and '=' not in value
        ):
            # Second special case is where a structure of a single
            # value whose member name is "Value" can be specified
            # as:
            # --instance-terminate-behavior shutdown
            #
            # gets parsed as:
            # {"Value": "shutdown"}
            return {'Value': value}

    def _should_parse_as_shorthand(self, cli_argument, value):
        # We first need to make sure this is a parameter that qualifies
        # for simplification.  The first short-circuit case is if it looks
        # like json we immediately return.
        if value and isinstance(value, list):
            check_val = value[0]
        else:
            check_val = value
        if isinstance(check_val, str) and check_val.strip().startswith(
            ('[', '{')
        ):
            LOG.debug(
                "Param %s looks like JSON, not considered for "
                "param shorthand.",
                cli_argument.py_name,
            )
            return False
        model = cli_argument.argument_model
        return _supports_shorthand_syntax(model)


class ParamShorthandDocGen(ParamShorthand):
    """Documentation generator for param shorthand syntax."""

    _DONT_DOC = object()
    _MAX_STACK = 3

    def supports_shorthand(self, argument_model):
        """Checks if a CLI argument supports shorthand syntax."""
        if argument_model is not None:
            return _supports_shorthand_syntax(argument_model)
        return False

    def generate_shorthand_example(
        self, cli_argument, command_name, operation_name
    ):
        """Generate documentation for a CLI argument.

        :type cli_argument: awscli.arguments.BaseCLIArgument
        :param cli_argument: The CLI argument which to generate
            documentation for.

        :return: Returns either a string or ``None``.  If a string
            is returned, it is the generated shorthand example.
            If a value of ``None`` is returned then this indicates
            that no shorthand syntax is available for the provided
            ``argument_model``.

        """
        docstring = self._handle_special_cases(
            cli_argument, command_name, operation_name
        )
        if docstring is self._DONT_DOC:
            return None
        elif docstring:
            return docstring

        # Otherwise we fall back to the normal docgen for shorthand
        # syntax.
        stack = []
        try:
            if cli_argument.argument_model.type_name == 'list':
                argument_model = cli_argument.argument_model.member
                return self._shorthand_docs(argument_model, stack) + ' ...'
            else:
                return self._shorthand_docs(cli_argument.argument_model, stack)
        except TooComplexError:
            return ''

    def _handle_special_cases(
        self, cli_argument, command_name, operation_name
    ):
        model = cli_argument.argument_model
        if (
            model.type_name == 'list'
            and model.member.type_name == 'structure'
            and len(model.member.members) == 1
            and self._uses_old_list_case(
                command_name, operation_name, cli_argument.name
            )
        ):
            member_name = list(model.member.members)[0]
            # Handle special case where the min/max is exactly one.
            metadata = model.metadata
            if metadata.get('min') == 1 and metadata.get('max') == 1:
                return '%s %s1' % (cli_argument.cli_name, member_name)
            return '%s %s1 %s2 %s3' % (
                cli_argument.cli_name,
                member_name,
                member_name,
                member_name,
            )
        elif (
            model.type_name == 'structure'
            and len(model.members) == 1
            and 'Value' in model.members
            and model.members['Value'].type_name == 'string'
        ):
            return self._DONT_DOC
        return ''

    def _shorthand_docs(self, argument_model, stack):
        if len(stack) > self._MAX_STACK:
            raise TooComplexError()
        if argument_model.type_name == 'structure':
            return self._structure_docs(argument_model, stack)
        elif argument_model.type_name == 'list':
            return self._list_docs(argument_model, stack)
        elif argument_model.type_name == 'map':
            return self._map_docs(argument_model, stack)
        else:
            return argument_model.type_name

    def _list_docs(self, argument_model, stack):
        list_member = argument_model.member
        stack.append(list_member.name)
        try:
            element_docs = self._shorthand_docs(argument_model.member, stack)
        finally:
            stack.pop()
        if list_member.type_name in COMPLEX_TYPES or len(stack) > 1:
            return '[%s,%s]' % (element_docs, element_docs)
        else:
            return '%s,%s' % (element_docs, element_docs)

    def _map_docs(self, argument_model, stack):
        k = argument_model.key
        stack.append(argument_model.value.name)
        try:
            value_docs = self._shorthand_docs(argument_model.value, stack)
        finally:
            stack.pop()
        start = 'KeyName1=%s,KeyName2=%s' % (value_docs, value_docs)
        if k.enum and not stack:
            start += '\n\nWhere valid key names are:\n'
            for enum in k.enum:
                start += '  %s\n' % enum
        elif stack:
            start = '{%s}' % start
        return start

    def _structure_docs(self, argument_model, stack):
        parts = []
        for name, member_shape in argument_model.members.items():
            if is_document_type_container(member_shape):
                continue
            parts.append(self._member_docs(name, member_shape, stack))
        inner_part = ','.join(parts)
        if not stack:
            return inner_part
        return '{%s}' % inner_part

    def _member_docs(self, name, shape, stack):
        if stack.count(shape.name) > 0:
            return '( ... recursive ... )'
        stack.append(shape.name)
        try:
            value_doc = self._shorthand_docs(shape, stack)
        finally:
            stack.pop()
        return '%s=%s' % (name, value_doc)



import re


class RegexFilter:
    def __init__(self, pattern, replacement):
        self._pattern = pattern
        self._replacement = replacement
        self._regex = None

    def filter_text(self, text):
        regex = self._get_regex()
        filtered_text = regex.subn(self._replacement, text)
        return filtered_text[0]

    def _get_regex(self):
        if self._regex is None:
            self._regex = re.compile(self._pattern)
        return self._regex



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.docs.shape import ShapeDocumenter
from botocore.docs.utils import py_type_name


class BaseParamsDocumenter(ShapeDocumenter):
    def document_params(self, section, shape, include=None, exclude=None):
        """Fills out the documentation for a section given a model shape.

        :param section: The section to write the documentation to.

        :param shape: The shape of the operation.

        :type include: Dictionary where keys are parameter names and
            values are the shapes of the parameter names.
        :param include: The parameter shapes to include in the documentation.

        :type exclude: List of the names of the parameters to exclude.
        :param exclude: The names of the parameters to exclude from
            documentation.
        """
        history = []
        self.traverse_and_document_shape(
            section=section,
            shape=shape,
            history=history,
            name=None,
            include=include,
            exclude=exclude,
        )

    def document_recursive_shape(self, section, shape, **kwargs):
        self._add_member_documentation(section, shape, **kwargs)

    def document_shape_default(
        self, section, shape, history, include=None, exclude=None, **kwargs
    ):
        self._add_member_documentation(section, shape, **kwargs)

    def document_shape_type_list(
        self, section, shape, history, include=None, exclude=None, **kwargs
    ):
        self._add_member_documentation(section, shape, **kwargs)
        param_shape = shape.member
        param_section = section.add_new_section(
            param_shape.name, context={'shape': shape.member.name}
        )
        self._start_nested_param(param_section)
        self.traverse_and_document_shape(
            section=param_section,
            shape=param_shape,
            history=history,
            name=None,
        )
        section = section.add_new_section('end-list')
        self._end_nested_param(section)

    def document_shape_type_map(
        self, section, shape, history, include=None, exclude=None, **kwargs
    ):
        self._add_member_documentation(section, shape, **kwargs)

        key_section = section.add_new_section(
            'key', context={'shape': shape.key.name}
        )
        self._start_nested_param(key_section)
        self._add_member_documentation(key_section, shape.key)

        param_section = section.add_new_section(
            shape.value.name, context={'shape': shape.value.name}
        )
        param_section.style.indent()
        self._start_nested_param(param_section)
        self.traverse_and_document_shape(
            section=param_section,
            shape=shape.value,
            history=history,
            name=None,
        )

        end_section = section.add_new_section('end-map')
        self._end_nested_param(end_section)
        self._end_nested_param(end_section)

    def document_shape_type_structure(
        self,
        section,
        shape,
        history,
        include=None,
        exclude=None,
        name=None,
        **kwargs,
    ):
        members = self._add_members_to_shape(shape.members, include)
        self._add_member_documentation(section, shape, name=name)
        for param in members:
            if exclude and param in exclude:
                continue
            param_shape = members[param]
            param_section = section.add_new_section(
                param, context={'shape': param_shape.name}
            )
            self._start_nested_param(param_section)
            self.traverse_and_document_shape(
                section=param_section,
                shape=param_shape,
                history=history,
                name=param,
            )
        section = section.add_new_section('end-structure')
        self._end_nested_param(section)

    def _add_member_documentation(self, section, shape, **kwargs):
        pass

    def _add_members_to_shape(self, members, include):
        if include:
            members = members.copy()
            for param in include:
                members[param.name] = param
        return members

    def _document_non_top_level_param_type(self, type_section, shape):
        special_py_type = self._get_special_py_type_name(shape)
        py_type = py_type_name(shape.type_name)

        type_format = '(%s) -- '
        if special_py_type is not None:
            # Special type can reference a linked class.
            # Italicizing it blows away the link.
            type_section.write(type_format % special_py_type)
        else:
            type_section.style.italics(type_format % py_type)

    def _start_nested_param(self, section):
        section.style.indent()
        section.style.new_line()

    def _end_nested_param(self, section):
        section.style.dedent()
        section.style.new_line()


class ResponseParamsDocumenter(BaseParamsDocumenter):
    """Generates the description for the response parameters"""

    EVENT_NAME = 'response-params'

    def _add_member_documentation(self, section, shape, name=None, **kwargs):
        name_section = section.add_new_section('param-name')
        name_section.write('- ')
        if name is not None:
            name_section.style.bold(f'{name} ')
        type_section = section.add_new_section('param-type')
        self._document_non_top_level_param_type(type_section, shape)

        documentation_section = section.add_new_section('param-documentation')
        if shape.documentation:
            documentation_section.style.indent()
            if getattr(shape, 'is_tagged_union', False):
                tagged_union_docs = section.add_new_section(
                    'param-tagged-union-docs'
                )
                note = (
                    '.. note::'
                    '    This is a Tagged Union structure. Only one of the '
                    '    following top level keys will be set: %s. '
                    '    If a client receives an unknown member it will '
                    '    set ``SDK_UNKNOWN_MEMBER`` as the top level key, '
                    '    which maps to the name or tag of the unknown '
                    '    member. The structure of ``SDK_UNKNOWN_MEMBER`` is '
                    '    as follows'
                )
                tagged_union_members_str = ', '.join(
                    [f'``{key}``' for key in shape.members.keys()]
                )
                unknown_code_example = (
                    '\'SDK_UNKNOWN_MEMBER\': '
                    '{\'name\': \'UnknownMemberName\'}'
                )
                tagged_union_docs.write(note % (tagged_union_members_str))
                example = section.add_new_section('param-unknown-example')
                example.style.codeblock(unknown_code_example)
            documentation_section.include_doc_string(shape.documentation)
        section.style.new_paragraph()

    def document_shape_type_event_stream(
        self, section, shape, history, **kwargs
    ):
        self.document_shape_type_structure(section, shape, history, **kwargs)


class RequestParamsDocumenter(BaseParamsDocumenter):
    """Generates the description for the request parameters"""

    EVENT_NAME = 'request-params'

    def document_shape_type_structure(
        self, section, shape, history, include=None, exclude=None, **kwargs
    ):
        if len(history) > 1:
            self._add_member_documentation(section, shape, **kwargs)
            section.style.indent()
        members = self._add_members_to_shape(shape.members, include)
        for i, param in enumerate(members):
            if exclude and param in exclude:
                continue
            param_shape = members[param]
            param_section = section.add_new_section(
                param, context={'shape': param_shape.name}
            )
            param_section.style.new_line()
            is_required = param in shape.required_members
            self.traverse_and_document_shape(
                section=param_section,
                shape=param_shape,
                history=history,
                name=param,
                is_required=is_required,
            )
        section = section.add_new_section('end-structure')
        if len(history) > 1:
            section.style.dedent()
        section.style.new_line()

    def _add_member_documentation(
        self,
        section,
        shape,
        name=None,
        is_top_level_param=False,
        is_required=False,
        **kwargs,
    ):
        py_type = self._get_special_py_type_name(shape)
        if py_type is None:
            py_type = py_type_name(shape.type_name)
        if is_top_level_param:
            type_section = section.add_new_section('param-type')
            type_section.write(f':type {name}: {py_type}')
            end_type_section = type_section.add_new_section('end-param-type')
            end_type_section.style.new_line()
            name_section = section.add_new_section('param-name')
            name_section.write(f':param {name}: ')

        else:
            name_section = section.add_new_section('param-name')
            name_section.write('- ')
            if name is not None:
                name_section.style.bold(f'{name} ')
            type_section = section.add_new_section('param-type')
            self._document_non_top_level_param_type(type_section, shape)

        if is_required:
            is_required_section = section.add_new_section('is-required')
            is_required_section.style.indent()
            is_required_section.style.bold('[REQUIRED] ')
        if shape.documentation:
            documentation_section = section.add_new_section(
                'param-documentation'
            )
            documentation_section.style.indent()
            if getattr(shape, 'is_tagged_union', False):
                tagged_union_docs = section.add_new_section(
                    'param-tagged-union-docs'
                )
                note = (
                    '.. note::'
                    '    This is a Tagged Union structure. Only one of the '
                    '    following top level keys can be set: %s. '
                )
                tagged_union_members_str = ', '.join(
                    [f'``{key}``' for key in shape.members.keys()]
                )
                tagged_union_docs.write(note % (tagged_union_members_str))
            documentation_section.include_doc_string(shape.documentation)
            self._add_special_trait_documentation(documentation_section, shape)
        end_param_section = section.add_new_section('end-param')
        end_param_section.style.new_paragraph()

    def _add_special_trait_documentation(self, section, shape):
        if 'idempotencyToken' in shape.metadata:
            self._append_idempotency_documentation(section)

    def _append_idempotency_documentation(self, section):
        docstring = 'This field is autopopulated if not provided.'
        section.write(docstring)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import sys

from botocore import xform_name
from botocore.stub import Stubber
from botocore.utils import ArgumentGenerator
from ruamel.yaml import YAML

from awscli.clidriver import CLIOperationCaller
from awscli.customizations.arguments import OverrideRequiredArgsArgument
from awscli.customizations.utils import get_shape_doc_overview
from awscli.utils import json_encoder


def register_generate_cli_skeleton(cli):
    cli.register('building-argument-table', add_generate_skeleton)


def add_generate_skeleton(session, operation_model, argument_table, **kwargs):
    # This argument cannot support operations with streaming output which
    # is designated by the argument name `outfile`.
    if 'outfile' not in argument_table:
        generate_cli_skeleton_argument = GenerateCliSkeletonArgument(
            session, operation_model
        )
        generate_cli_skeleton_argument.add_to_arg_table(argument_table)


class GenerateCliSkeletonArgument(OverrideRequiredArgsArgument):
    """This argument writes a generated JSON skeleton to stdout

    The argument, if present in the command line, will prevent the intended
    command from taking place. Instead, it will generate a JSON skeleton and
    print it to standard output.
    """

    ARG_DATA = {
        'name': 'generate-cli-skeleton',
        'help_text': (
            'Prints a JSON skeleton to standard output without sending '
            'an API request. If provided with no value or the value '
            '``input``, prints a sample input JSON that can be used as an '
            'argument for ``--cli-input-json``. Similarly, if provided '
            '``yaml-input`` it will print a sample input YAML that can be '
            'used with ``--cli-input-yaml``. If provided with the value '
            '``output``, it validates the command inputs and returns a '
            'sample output JSON for that command. '
            'The generated JSON skeleton is not stable between versions '
            'of the AWS CLI and there are no backwards compatibility '
            'guarantees in the JSON skeleton generated.'
        ),
        'nargs': '?',
        'const': 'input',
        'choices': ['input', 'output', 'yaml-input'],
    }

    def __init__(self, session, operation_model):
        super(GenerateCliSkeletonArgument, self).__init__(session)
        self._operation_model = operation_model

    def _register_argument_action(self):
        self._session.register('calling-command.*', self.generate_skeleton)
        super(GenerateCliSkeletonArgument, self)._register_argument_action()

    def override_required_args(self, argument_table, args, **kwargs):
        arg_name = '--' + self.name
        if arg_name in args:
            arg_location = args.index(arg_name)
            try:
                # If the value of --generate-cli-skeleton is ``output``,
                # do not force required arguments to be optional as
                # ``--generate-cli-skeleton output`` validates commands
                # as well as print out the sample output.
                if args[arg_location + 1] == 'output':
                    return
            except IndexError:
                pass
            super(GenerateCliSkeletonArgument, self).override_required_args(
                argument_table, args, **kwargs
            )

    def generate_skeleton(
        self, call_parameters, parsed_args, parsed_globals, **kwargs
    ):
        if not getattr(parsed_args, 'generate_cli_skeleton', None):
            return
        arg_value = parsed_args.generate_cli_skeleton
        return getattr(
            self, '_generate_%s_skeleton' % arg_value.replace('-', '_')
        )(call_parameters=call_parameters, parsed_globals=parsed_globals)

    def _generate_yaml_input_skeleton(self, **kwargs):
        input_shape = self._operation_model.input_shape
        yaml = YAML()
        yaml.representer.add_representer(_Bytes, _Bytes.represent)
        skeleton = yaml.map()
        if input_shape is not None:
            argument_generator = YAMLArgumentGenerator()
            skeleton = argument_generator.generate_skeleton(input_shape)
        yaml.dump(skeleton, sys.stdout)
        return 0

    def _generate_input_skeleton(self, **kwargs):
        outfile = sys.stdout
        input_shape = self._operation_model.input_shape
        skeleton = {}
        if input_shape is not None:
            argument_generator = ArgumentGenerator()
            skeleton = argument_generator.generate_skeleton(input_shape)
        json.dump(skeleton, outfile, indent=4, default=json_encoder)
        outfile.write('\n')
        return 0

    def _generate_output_skeleton(
        self, call_parameters, parsed_globals, **kwargs
    ):
        service_name = self._operation_model.service_model.service_name
        operation_name = self._operation_model.name
        return StubbedCLIOperationCaller(self._session).invoke(
            service_name, operation_name, call_parameters, parsed_globals
        )


class StubbedCLIOperationCaller(CLIOperationCaller):
    """A stubbed CLIOperationCaller

    It generates a fake response and uses the response and provided parameters
    to make a stubbed client call for an operation command.
    """

    def _make_client_call(
        self, client, operation_name, parameters, parsed_globals
    ):
        method_name = xform_name(operation_name)
        operation_model = client.meta.service_model.operation_model(
            operation_name
        )
        fake_response = {}
        if operation_model.output_shape:
            argument_generator = ArgumentGenerator(use_member_names=True)
            fake_response = argument_generator.generate_skeleton(
                operation_model.output_shape
            )
        with Stubber(client) as stubber:
            stubber.add_response(method_name, fake_response)
            return getattr(client, method_name)(**parameters)


class _Bytes:
    @classmethod
    def represent(cls, dumper, data):
        return dumper.represent_scalar('tag:yaml.org,2002:binary', '')


class YAMLArgumentGenerator(ArgumentGenerator):
    def __init__(self, use_member_names=False, yaml=None):
        super(YAMLArgumentGenerator, self).__init__(
            use_member_names=use_member_names
        )
        self._yaml = yaml
        if self._yaml is None:
            self._yaml = YAML()

    def _generate_skeleton(self, shape, stack, name=''):
        # YAML supports binary, so add in boilerplate for that instead of
        # putting in None. Here were' using a custom type so that we can ensure
        # we serialize it correctly on python 2 and to make the
        # serialization output more usable on python 3.
        if shape.type_name == 'blob':
            return _Bytes()
        return super(YAMLArgumentGenerator, self)._generate_skeleton(
            shape, stack, name
        )

    def _generate_type_structure(self, shape, stack):
        if stack.count(shape.name) > 1:
            return self._yaml.map()
        skeleton = self._yaml.map()
        for member_name, member_shape in shape.members.items():
            skeleton[member_name] = self._generate_skeleton(
                member_shape, stack, name=member_name
            )
            is_required = member_name in shape.required_members
            self._add_member_comments(
                skeleton, member_name, member_shape, is_required
            )
        return skeleton

    def _add_member_comments(
        self, skeleton, member_name, member_shape, is_required
    ):
        comment_components = []
        if is_required:
            comment_components.append('[REQUIRED]')
        comment_components.append(get_shape_doc_overview(member_shape))
        if getattr(member_shape, 'enum', None):
            comment_components.append(
                self._get_enums_comment_content(member_shape.enum)
            )
        comment = ' '.join(comment_components)
        if comment and not comment.isspace():
            skeleton.yaml_add_eol_comment('# ' + comment, member_name)

    def _get_enums_comment_content(self, enums):
        return 'Valid values are: %s.' % ', '.join(enums)

    def _generate_type_map(self, shape, stack):
        # YAML has support for ordered maps, so don't use ordereddicts
        # because that isn't necessary and it makes the output harder to
        # understand and read.
        return dict(
            super(YAMLArgumentGenerator, self)._generate_type_map(shape, stack)
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import webbrowser

import botocore
from awscrt.crypto import EC, ECType
from botocore.exceptions import ProfileNotFound
from botocore.loaders import Loader
from botocore.utils import (
    JSONFileCache,
    LoginCredentialsLoader,
    get_login_token_cache_directory,
)

from awscli.compat import compat_input
from awscli.customizations.commands import BasicCommand
from awscli.customizations.configure.sso import (
    PTKPrompt,
    RequiredInputValidator,
)
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.exceptions import ConfigurationError
from awscli.customizations.login.utils import (
    CrossDeviceLoginTokenFetcher,
    LoginType,
    SameDeviceLoginTokenFetcher,
)
from awscli.customizations.sso.utils import (
    AuthCodeFetcher,
    OpenBrowserHandler,
    PrintOnlyHandler,
    open_browser_with_original_ld_path,
)
from awscli.customizations.utils import uni_print

DEFAULT_REGION_FOR_PROMPT = 'us-east-1'


class LoginCommand(BasicCommand):
    NAME = 'login'
    DESCRIPTION = (
        'Login for local development using AWS Management Console '
        'credentials. Each time the ``login`` command is called, the '
        'CLI will acquire temporary credentials and a refresh token '
        'that correspond to your selected console session. '
        'The CLI will refresh the temporary credentials automatically '
        'as long as the refresh token is valid.\n\n'
        'You can override the directory where the CLI will store the '
        'temporary credentials with the ``AWS_LOGIN_CACHE_DIRECTORY`` '
        'environment variable.'
    )
    EXAMPLES = BasicCommand.FROM_FILE()
    ARG_TABLE = [
        {
            'name': 'remote',
            'action': 'store_true',
            'default': False,
            'help_text': (
                'Disables the local callback server and redirect-based auth '
                'flow. Instead displays the URL and prompts you to paste the '
                'authorization code that is displayed after logging in. This is '
                'intended when running the CLI on remote hosts via SSH '
                'where a local browser is not available.'
            ),
        }
    ]

    def __init__(
        self,
        session,
        token_loader=None,
        config_file_writer=None,
    ):
        super().__init__(session)

        # Records if we had to prompt the user for a region,
        # which we'll save to the profile when we do
        self._prompted_for_region = False

        if token_loader is None:
            token_cache = JSONFileCache(get_login_token_cache_directory())
            token_loader = LoginCredentialsLoader(token_cache)
        self._token_loader = token_loader

        if config_file_writer is None:
            config_file_writer = ConfigFileWriter()
        self._config_file_writer = config_file_writer

    def _run_main(self, parsed_args, parsed_globals):
        region = self._resolve_region(parsed_globals)
        profile_name = self.resolve_profile_name()
        sign_in_type = self.resolve_sign_in_type(parsed_args)

        # If the profile specified via --profile doesn't already exist
        # add it to the session so the client creation still succeeds.
        # If the login is successful we'll save the profile at the end.
        if profile_name not in self._session.available_profiles:
            self._session._profile_map[profile_name] = {}

        # Abort if the profile is already configured with a different style
        # of credentials, since they'd still have precedence over login
        self.ensure_profile_does_not_have_existing_credentials(profile_name)

        config = botocore.config.Config(
            region_name=region,
            signature_version=botocore.UNSIGNED,
        )
        client = self._session.create_client(
            'signin',
            config=config,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )

        private_key = EC.new_generate(ECType.P_256)

        if sign_in_type is LoginType.SAME_DEVICE:
            token_fetcher = SameDeviceLoginTokenFetcher(
                client=client,
                auth_code_fetcher=AuthCodeFetcher(),
                on_pending_authorization=OpenBrowserHandler(
                    open_browser=open_browser_with_original_ld_path
                ),
                private_key=private_key,
            )
        else:  # Cross-Device
            token_fetcher = CrossDeviceLoginTokenFetcher(
                client=client,
                on_pending_authorization=PrintOnlyHandler(),
                private_key=private_key,
            )

        # Execute the browser flow, and the initial call to the token endpoint
        access_token, session_id = token_fetcher.fetch_token()

        if not self.accept_change_to_existing_profile_if_needed(
            profile_name, session_id
        ):
            return

        # Cache the access token to disk
        self._token_loader.save_token(session_id, access_token)

        # Update the specified profile with the session (and region if not set)
        self._update_profile_with_login_session(
            profile_name, session_id, region
        )

        uni_print(
            f'\nUpdated profile {profile_name} to use {session_id} credentials.\n'
        )
        if profile_name != 'default':
            uni_print(
                f'Use "--profile {profile_name}" to use the new credentials, '
                f'such as "aws sts get-caller-identity --profile {profile_name}"\n'
            )

    def accept_change_to_existing_profile_if_needed(
        self, profile_name, new_session_id
    ):
        config = self._session.full_config['profiles'].get(profile_name, {})

        if 'login_session' not in config:
            return True

        existing_session_id = config['login_session']

        if existing_session_id == new_session_id:
            return True

        while True:
            response = compat_input(
                f'\nProfile {profile_name} is already configured to use session '
                f'{existing_session_id}. Do you want to overwrite it to use '
                f'{new_session_id} instead? (y/n): '
            )

            if response.lower() in ('y', 'yes'):
                return True
            elif response.lower() in ('n', 'no'):
                return False
            else:
                uni_print('Invalid response. Please enter "y" or "n"')

    def ensure_profile_does_not_have_existing_credentials(self, profile_name):
        """
        Raises an error if the specified profile is already
        configured with a different style of credentials.
        """
        config = self._session.full_config['profiles'].get(profile_name, {})
        existing_credentials_style = None

        if 'web_identity_token_file' in config:
            existing_credentials_style = 'Web Identity'
        elif 'sso_role_name' in config or 'sso_account_id' in config:
            existing_credentials_style = 'SSO'
        elif 'aws_access_key_id' in config:
            existing_credentials_style = 'Access Key'
        elif 'role_arn' in config:
            existing_credentials_style = 'Assume Role'
        elif 'credential_process' in config:
            existing_credentials_style = 'Credential Process'

        if existing_credentials_style:
            raise ConfigurationError(
                f'Profile \'{profile_name}\' is already configured '
                f'with {existing_credentials_style} credentials.\n\n'
                f'You may run \'aws login --profile new-profile-name\' to '
                f'create a new profile with the specified name. Otherwise you '
                f'must first manually remove the existing credentials '
                f'from \'{profile_name}\'.\n'
            )

        return False

    @staticmethod
    def resolve_sign_in_type(parsed_args):
        if parsed_args.remote:
            return LoginType.CROSS_DEVICE

        return LoginType.SAME_DEVICE

    def resolve_profile_name(self):
        profile = self._session.profile
        return profile if profile is not None else 'default'

    def _resolve_region(self, parsed_globals):
        if parsed_globals.region:
            return parsed_globals.region
        try:
            if self._session.get_config_variable('region'):
                return self._session.get_config_variable('region')
        except ProfileNotFound:
            pass

        return self._prompt_for_region()

    def _prompt_for_region(self):
        prompter = PTKPrompt()
        self._prompted_for_region = True
        uni_print(
            'No AWS region has been configured. '
            'The AWS region is the geographic location of your AWS resources.\n\n'
            'If you have used AWS before and already have resources in your '
            'account, specify which region they were created in. '
            'If you have not created resources in your account before, you '
            'can pick the region closest to you: '
            'https://docs.aws.amazon.com/global-infrastructure/latest/regions/aws-regions.html.\n\n'
            'You are able to change the region in the CLI at any time with '
            'the command "aws configure set region NEW_REGION".\n'
        )
        new_region = prompter.get_value(
            DEFAULT_REGION_FOR_PROMPT,
            'AWS Region',
            completions=list(self._load_all_regions()),
            validator=RequiredInputValidator(DEFAULT_REGION_FOR_PROMPT),
        )

        return new_region if new_region else DEFAULT_REGION_FOR_PROMPT

    @staticmethod
    def _load_all_regions():
        """Loads all regions from all partitions via partitions.json"""
        loader = Loader()
        for partition in loader.load_data('partitions')['partitions']:
            yield from partition['regions'].keys()

    def _update_profile_with_login_session(
        self, profile_name, session_id, region
    ):
        config_filename = os.path.expanduser(
            self._session.get_config_variable('config_file')
        )

        section_name = (
            'default'
            if profile_name == 'default'
            else f'profile {profile_name}'
        )

        new_values = {
            '__section__': section_name,
            'login_session': session_id,
        }

        if self._prompted_for_region:
            new_values['region'] = region

        self._config_file_writer.update_config(new_values, config_filename)
